@@ -13,7 +13,7 @@
 // All changes made under the Poppler project to this file are licensed
 // under GPL version 2 or later
 //
-// Copyright (C) 2005, 2006, 2008-2010, 2012, 2014, 2015, 2017-2023 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2005, 2006, 2008-2010, 2012, 2014, 2015, 2017-2024 Albert Astals Cid <aacid@kde.org>
 // Copyright (C) 2005, 2006 Kristian Høgsberg <krh@redhat.com>
 // Copyright (C) 2006 Takashi Iwai <tiwai@suse.de>
 // Copyright (C) 2007 Julien Rebetez <julienr@svn.gnome.org>
@@ -38,6 +38,8 @@
 // Copyright (C) 2021, 2022, 2024 Oliver Sander <oliver.sander@tu-dresden.de>
 // Copyright (C) 2023 Khaled Hosny <khaled@aliftype.com>
 // Copyright (C) 2024 Nelson Benítez León <nbenitezl@gmail.com>
+// Copyright (C) 2024 g10 Code GmbH, Author: Sune Stolborg Vuorela <sune@vuorela.dk>
+// Copyright (C) 2024 Vincent Lefevre <vincent@vinc17.net>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -184,17 +186,6 @@ GfxFontLoc::GfxFontLoc(GfxFontLoc &&other) noexcept = default;
 
 GfxFontLoc &GfxFontLoc::operator=(GfxFontLoc &&other) noexcept = default;
 
-void GfxFontLoc::setPath(GooString *pathA)
-{
-    path = pathA->toStr();
-    delete pathA;
-}
-
-const GooString *GfxFontLoc::pathAsGooString() const
-{
-    return (const GooString *)(&path);
-}
-
 //------------------------------------------------------------------------
 // GfxFont
 //------------------------------------------------------------------------
@@ -608,31 +599,30 @@ void GfxFont::readFontDescriptor(XRef *xref, Dict *fontDict)
     }
 }
 
-CharCodeToUnicode *GfxFont::readToUnicodeCMap(Dict *fontDict, int nBits, CharCodeToUnicode *ctu)
+std::unique_ptr<CharCodeToUnicode> GfxFont::readToUnicodeCMap(Dict *fontDict, int nBits, std::unique_ptr<CharCodeToUnicode> ctu)
 {
-    GooString *buf;
 
     Object obj1 = fontDict->lookup("ToUnicode");
     if (!obj1.isStream()) {
-        return nullptr;
+        return ctu;
     }
-    buf = new GooString();
-    obj1.getStream()->fillGooString(buf);
+    auto buf = std::make_unique<GooString>();
+    obj1.getStream()->fillGooString(buf.get());
     obj1.streamClose();
     if (ctu) {
-        ctu->mergeCMap(buf, nBits);
+        ctu->mergeCMap(buf.get(), nBits);
     } else {
-        ctu = CharCodeToUnicode::parseCMap(buf, nBits);
+        ctu = CharCodeToUnicode::parseCMap(buf.get(), nBits);
     }
     hasToUnicode = true;
-    delete buf;
     return ctu;
 }
 
 std::optional<GfxFontLoc> GfxFont::locateFont(XRef *xref, PSOutputDev *ps, GooString *substituteFontName)
 {
     SysFontType sysFontType;
-    GooString *path, *base14Name;
+    std::optional<std::string> path;
+    GooString *base14Name;
     int substIdx, fontNum;
     bool embed;
 
@@ -688,7 +678,7 @@ std::optional<GfxFontLoc> GfxFont::locateFont(XRef *xref, PSOutputDev *ps, GooSt
         GfxFontLoc fontLoc;
         fontLoc.locType = gfxFontLocResident;
         fontLoc.fontType = fontType1;
-        fontLoc.path = *name;
+        fontLoc.path = name.value_or(std::string());
         return fontLoc;
     }
 
@@ -703,7 +693,7 @@ std::optional<GfxFontLoc> GfxFont::locateFont(XRef *xref, PSOutputDev *ps, GooSt
 
     //----- external font file (fontFile, fontDir)
     if (name && (path = globalParams->findFontFile(*name))) {
-        if (std::optional<GfxFontLoc> fontLoc = getExternalFont(path, isCIDFont())) {
+        if (std::optional<GfxFontLoc> fontLoc = getExternalFont(*path, isCIDFont())) {
             return fontLoc;
         }
     }
@@ -712,7 +702,7 @@ std::optional<GfxFontLoc> GfxFont::locateFont(XRef *xref, PSOutputDev *ps, GooSt
     if (!ps && !isCIDFont() && ((Gfx8BitFont *)this)->base14) {
         base14Name = new GooString(((Gfx8BitFont *)this)->base14->base14Name);
         if ((path = globalParams->findBase14FontFile(base14Name, this, substituteFontName))) {
-            if (std::optional<GfxFontLoc> fontLoc = getExternalFont(path, false)) {
+            if (std::optional<GfxFontLoc> fontLoc = getExternalFont(*path, false)) {
                 delete base14Name;
                 return fontLoc;
             }
@@ -727,13 +717,13 @@ std::optional<GfxFontLoc> GfxFont::locateFont(XRef *xref, PSOutputDev *ps, GooSt
                 GfxFontLoc fontLoc;
                 fontLoc.locType = gfxFontLocExternal;
                 fontLoc.fontType = fontCIDType2;
-                fontLoc.setPath(path);
+                fontLoc.path = *path;
                 fontLoc.fontNum = fontNum;
                 return fontLoc;
             }
         } else {
             GfxFontLoc fontLoc;
-            fontLoc.setPath(path);
+            fontLoc.path = *path;
             fontLoc.locType = gfxFontLocExternal;
             if (sysFontType == sysFontTTF || sysFontType == sysFontTTC) {
                 fontLoc.fontType = fontTrueType;
@@ -743,7 +733,6 @@ std::optional<GfxFontLoc> GfxFont::locateFont(XRef *xref, PSOutputDev *ps, GooSt
             }
             return fontLoc;
         }
-        delete path;
     }
 
     if (!isCIDFont()) {
@@ -774,7 +763,7 @@ std::optional<GfxFontLoc> GfxFont::locateFont(XRef *xref, PSOutputDev *ps, GooSt
         } else {
             path = globalParams->findFontFile(substName);
             if (path) {
-                if (std::optional<GfxFontLoc> fontLoc = getExternalFont(path, false)) {
+                if (std::optional<GfxFontLoc> fontLoc = getExternalFont(*path, false)) {
                     error(errSyntaxWarning, -1, "Substituting font '{0:s}' for '{1:s}'", base14SubstFonts[substIdx], name ? name->c_str() : "");
                     name = base14SubstFonts[substIdx];
                     fontLoc->substIdx = substIdx;
@@ -791,12 +780,12 @@ std::optional<GfxFontLoc> GfxFont::locateFont(XRef *xref, PSOutputDev *ps, GooSt
     return std::nullopt;
 }
 
-std::optional<GfxFontLoc> GfxFont::getExternalFont(GooString *path, bool cid)
+std::optional<GfxFontLoc> GfxFont::getExternalFont(const std::string &path, bool cid)
 {
     FoFiIdentifierType fft;
     GfxFontType fontType;
 
-    fft = FoFiIdentifier::identifyFile(path->c_str());
+    fft = FoFiIdentifier::identifyFile(path.c_str());
     switch (fft) {
     case fofiIdType1PFA:
     case fofiIdType1PFB:
@@ -825,13 +814,12 @@ std::optional<GfxFontLoc> GfxFont::getExternalFont(GooString *path, bool cid)
         break;
     }
     if (fontType == fontUnknownType || (cid ? (fontType < fontCIDType0) : (fontType >= fontCIDType0))) {
-        delete path;
         return std::nullopt;
     }
     GfxFontLoc fontLoc;
     fontLoc.locType = gfxFontLocExternal;
     fontLoc.fontType = fontType;
-    fontLoc.setPath(path);
+    fontLoc.path = path;
     return fontLoc;
 }
 
@@ -982,8 +970,6 @@ Gfx8BitFont::Gfx8BitFont(XRef *xref, const char *tagA, Ref idA, std::optional<st
     Object obj1;
     int n, a, b, m;
 
-    ctu = nullptr;
-
     // do font name substitution for various aliases of the Base 14 font
     // names
     base14 = nullptr;
@@ -1324,7 +1310,7 @@ Gfx8BitFont::Gfx8BitFont(XRef *xref, const char *tagA, Ref idA, std::optional<st
     // existing entries in ctu, i.e., the ToUnicode CMap takes
     // precedence, but the other encoding info is allowed to fill in any
     // holes
-    readToUnicodeCMap(fontDict, 16, ctu);
+    ctu = readToUnicodeCMap(fontDict, 16, std::move(ctu));
 
     //----- get the character widths -----
 
@@ -1424,7 +1410,6 @@ Gfx8BitFont::~Gfx8BitFont()
             gfree(enc[i]);
         }
     }
-    ctu->decRefCnt();
 }
 
 // This function is in part a derived work of the Adobe Glyph Mapping
@@ -1555,7 +1540,7 @@ int Gfx8BitFont::getNextChar(const char *s, int len, CharCode *code, Unicode con
 
 const CharCodeToUnicode *Gfx8BitFont::getToUnicode() const
 {
-    return ctu;
+    return ctu.get();
 }
 
 int *Gfx8BitFont::getCodeToGIDMap(FoFiTrueType *ff)
@@ -1733,21 +1718,15 @@ GfxCIDFont::GfxCIDFont(XRef *xref, const char *tagA, Ref idA, std::optional<std:
     Object desFontDictObj;
     Object obj1, obj2, obj3, obj4, obj5, obj6;
     int c1, c2;
-    int excepsSize;
 
     ascent = 0.95;
     descent = -0.35;
     fontBBox[0] = fontBBox[1] = fontBBox[2] = fontBBox[3] = 0;
     collection = nullptr;
-    ctu = nullptr;
     ctuUsesCharCode = true;
     widths.defWidth = 1.0;
     widths.defHeight = -1.0;
     widths.defVY = 0.880;
-    widths.exceps = nullptr;
-    widths.nExceps = 0;
-    widths.excepsV = nullptr;
-    widths.nExcepsV = 0;
     cidToGID = nullptr;
     cidToGIDLen = 0;
 
@@ -1863,7 +1842,6 @@ GfxCIDFont::GfxCIDFont(XRef *xref, const char *tagA, Ref idA, std::optional<std:
     // char width exceptions
     obj1 = desFontDict->lookup("W");
     if (obj1.isArray()) {
-        excepsSize = 0;
         int i = 0;
         while (i + 1 < obj1.arrayGetLength()) {
             obj2 = obj1.arrayGet(i);
@@ -1871,33 +1849,21 @@ GfxCIDFont::GfxCIDFont(XRef *xref, const char *tagA, Ref idA, std::optional<std:
             if (obj2.isInt() && obj3.isInt() && i + 2 < obj1.arrayGetLength()) {
                 obj4 = obj1.arrayGet(i + 2);
                 if (obj4.isNum()) {
-                    if (widths.nExceps == excepsSize) {
-                        excepsSize += 16;
-                        widths.exceps = (GfxFontCIDWidthExcep *)greallocn(widths.exceps, excepsSize, sizeof(GfxFontCIDWidthExcep));
-                    }
-                    widths.exceps[widths.nExceps].first = obj2.getInt();
-                    widths.exceps[widths.nExceps].last = obj3.getInt();
-                    widths.exceps[widths.nExceps].width = obj4.getNum() * 0.001;
-                    ++widths.nExceps;
+                    GfxFontCIDWidthExcep excep { static_cast<CID>(obj2.getInt()), static_cast<CID>(obj3.getInt()), obj4.getNum() * 0.001 };
+                    widths.exceps.push_back(excep);
                 } else {
                     error(errSyntaxError, -1, "Bad widths array in Type 0 font");
                 }
                 i += 3;
             } else if (obj2.isInt() && obj3.isArray()) {
-                if (widths.nExceps + obj3.arrayGetLength() > excepsSize) {
-                    excepsSize = (widths.nExceps + obj3.arrayGetLength() + 15) & ~15;
-                    widths.exceps = (GfxFontCIDWidthExcep *)greallocn(widths.exceps, excepsSize, sizeof(GfxFontCIDWidthExcep));
-                }
                 int j = obj2.getInt();
                 if (likely(j < INT_MAX - obj3.arrayGetLength())) {
                     for (int k = 0; k < obj3.arrayGetLength(); ++k) {
                         obj4 = obj3.arrayGet(k);
                         if (obj4.isNum()) {
-                            widths.exceps[widths.nExceps].first = j;
-                            widths.exceps[widths.nExceps].last = j;
-                            widths.exceps[widths.nExceps].width = obj4.getNum() * 0.001;
+                            GfxFontCIDWidthExcep excep { static_cast<CID>(j), static_cast<CID>(j), obj4.getNum() * 0.001 };
+                            widths.exceps.push_back(excep);
                             ++j;
-                            ++widths.nExceps;
                         } else {
                             error(errSyntaxError, -1, "Bad widths array in Type 0 font");
                         }
@@ -1909,7 +1875,7 @@ GfxCIDFont::GfxCIDFont(XRef *xref, const char *tagA, Ref idA, std::optional<std:
                 ++i;
             }
         }
-        std::sort(widths.exceps, widths.exceps + widths.nExceps, cmpWidthExcepFunctor());
+        std::sort(widths.exceps.begin(), widths.exceps.end(), cmpWidthExcepFunctor());
     }
 
     // default metrics for vertical font
@@ -1928,42 +1894,25 @@ GfxCIDFont::GfxCIDFont(XRef *xref, const char *tagA, Ref idA, std::optional<std:
     // char metric exceptions for vertical font
     obj1 = desFontDict->lookup("W2");
     if (obj1.isArray()) {
-        excepsSize = 0;
         int i = 0;
         while (i + 1 < obj1.arrayGetLength()) {
             obj2 = obj1.arrayGet(i);
             obj3 = obj1.arrayGet(i + 1);
             if (obj2.isInt() && obj3.isInt() && i + 4 < obj1.arrayGetLength()) {
                 if ((obj4 = obj1.arrayGet(i + 2), obj4.isNum()) && (obj5 = obj1.arrayGet(i + 3), obj5.isNum()) && (obj6 = obj1.arrayGet(i + 4), obj6.isNum())) {
-                    if (widths.nExcepsV == excepsSize) {
-                        excepsSize += 16;
-                        widths.excepsV = (GfxFontCIDWidthExcepV *)greallocn(widths.excepsV, excepsSize, sizeof(GfxFontCIDWidthExcepV));
-                    }
-                    widths.excepsV[widths.nExcepsV].first = obj2.getInt();
-                    widths.excepsV[widths.nExcepsV].last = obj3.getInt();
-                    widths.excepsV[widths.nExcepsV].height = obj4.getNum() * 0.001;
-                    widths.excepsV[widths.nExcepsV].vx = obj5.getNum() * 0.001;
-                    widths.excepsV[widths.nExcepsV].vy = obj6.getNum() * 0.001;
-                    ++widths.nExcepsV;
+                    GfxFontCIDWidthExcepV excepV { static_cast<CID>(obj2.getInt()), static_cast<CID>(obj3.getInt()), obj4.getNum() * 0.001, obj5.getNum() * 0.001, obj6.getNum() * 0.001 };
+                    widths.excepsV.push_back(excepV);
                 } else {
                     error(errSyntaxError, -1, "Bad widths (W2) array in Type 0 font");
                 }
                 i += 5;
             } else if (obj2.isInt() && obj3.isArray()) {
-                if (widths.nExcepsV + obj3.arrayGetLength() / 3 > excepsSize) {
-                    excepsSize = (widths.nExcepsV + obj3.arrayGetLength() / 3 + 15) & ~15;
-                    widths.excepsV = (GfxFontCIDWidthExcepV *)greallocn(widths.excepsV, excepsSize, sizeof(GfxFontCIDWidthExcepV));
-                }
                 int j = obj2.getInt();
                 for (int k = 0; k < obj3.arrayGetLength(); k += 3) {
                     if ((obj4 = obj3.arrayGet(k), obj4.isNum()) && (obj5 = obj3.arrayGet(k + 1), obj5.isNum()) && (obj6 = obj3.arrayGet(k + 2), obj6.isNum())) {
-                        widths.excepsV[widths.nExcepsV].first = j;
-                        widths.excepsV[widths.nExcepsV].last = j;
-                        widths.excepsV[widths.nExcepsV].height = obj4.getNum() * 0.001;
-                        widths.excepsV[widths.nExcepsV].vx = obj5.getNum() * 0.001;
-                        widths.excepsV[widths.nExcepsV].vy = obj6.getNum() * 0.001;
+                        GfxFontCIDWidthExcepV excepV { static_cast<CID>(j), static_cast<CID>(j), obj4.getNum() * 0.001, obj5.getNum() * 0.001, obj6.getNum() * 0.001 };
+                        widths.excepsV.push_back(excepV);
                         ++j;
-                        ++widths.nExcepsV;
                     } else {
                         error(errSyntaxError, -1, "Bad widths (W2) array in Type 0 font");
                     }
@@ -1974,7 +1923,7 @@ GfxCIDFont::GfxCIDFont(XRef *xref, const char *tagA, Ref idA, std::optional<std:
                 ++i;
             }
         }
-        std::sort(widths.excepsV, widths.excepsV + widths.nExcepsV, cmpWidthExcepVFunctor());
+        std::sort(widths.excepsV.begin(), widths.excepsV.end(), cmpWidthExcepVFunctor());
     }
 
     ok = true;
@@ -1985,11 +1934,6 @@ GfxCIDFont::~GfxCIDFont()
     if (collection) {
         delete collection;
     }
-    if (ctu) {
-        ctu->decRefCnt();
-    }
-    gfree(widths.exceps);
-    gfree(widths.excepsV);
     if (cidToGID) {
         gfree(cidToGID);
     }
@@ -2036,9 +1980,9 @@ int GfxCIDFont::getNextChar(const char *s, int len, CharCode *code, Unicode cons
         h = widths.defHeight;
         vx = getWidth(cid) / 2;
         vy = widths.defVY;
-        if (widths.nExcepsV > 0 && cid >= widths.excepsV[0].first) {
+        if (widths.excepsV.size() > 0 && cid >= widths.excepsV[0].first) {
             a = 0;
-            b = widths.nExcepsV;
+            b = widths.excepsV.size();
             // invariant: widths.excepsV[a].first <= cid < widths.excepsV[b].first
             while (b - a > 1) {
                 m = (a + b) / 2;
@@ -2071,7 +2015,7 @@ int GfxCIDFont::getWMode() const
 
 const CharCodeToUnicode *GfxCIDFont::getToUnicode() const
 {
-    return ctu;
+    return ctu.get();
 }
 
 const GooString *GfxCIDFont::getCollection() const
@@ -2206,10 +2150,9 @@ int *GfxCIDFont::getCodeToGIDMap(FoFiTrueType *ff, int *codeToGIDLen)
     humap = new Unicode[n * N_UCS_CANDIDATES];
     memset(humap, 0, sizeof(Unicode) * n * N_UCS_CANDIDATES);
     if (lp->collection != nullptr) {
-        CharCodeToUnicode *tctu;
         GooString tname(lp->toUnicodeMap);
 
-        if ((tctu = CharCodeToUnicode::parseCMapFromFile(&tname, 16)) != nullptr) {
+        if (std::unique_ptr<CharCodeToUnicode> tctu = CharCodeToUnicode::parseCMapFromFile(&tname, 16)) {
             tumap = new Unicode[n];
             CharCode cid;
             for (cid = 0; cid < n; cid++) {
@@ -2224,7 +2167,6 @@ int *GfxCIDFont::getCodeToGIDMap(FoFiTrueType *ff, int *codeToGIDLen)
                     tumap[cid] = 0;
                 }
             }
-            delete tctu;
         }
         vumap = new Unicode[n];
         memset(vumap, 0, sizeof(Unicode) * n);
@@ -2245,7 +2187,7 @@ int *GfxCIDFont::getCodeToGIDMap(FoFiTrueType *ff, int *codeToGIDLen)
         if (getCollection()->cmp("Adobe-Identity") == 0) {
             error(errSyntaxError, -1, "non-embedded font using identity encoding: {0:s}", name ? name->c_str() : "(null)");
         } else {
-            error(errSyntaxError, -1, "Unknown character collection {0:t}\n", getCollection());
+            error(errSyntaxError, -1, "Unknown character collection {0:t}", getCollection());
         }
         if (ctu) {
             CharCode cid;
@@ -2331,9 +2273,9 @@ double GfxCIDFont::getWidth(CID cid) const
     int a, b, m;
 
     w = widths.defWidth;
-    if (widths.nExceps > 0 && cid >= widths.exceps[0].first) {
+    if (widths.exceps.size() > 0 && cid >= widths.exceps[0].first) {
         a = 0;
-        b = widths.nExceps;
+        b = widths.exceps.size();
         // invariant: widths.exceps[a].first <= cid < widths.exceps[b].first
         while (b - a > 1) {
             m = (a + b) / 2;
