@@ -15,7 +15,7 @@
 //
 // Copyright (C) 2005 Martin Kretzschmar <martink@gnome.org>
 // Copyright (C) 2005, 2006 Kristian Høgsberg <krh@redhat.com>
-// Copyright (C) 2005, 2007-2010, 2012, 2015, 2017-2023 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2005, 2007-2010, 2012, 2015, 2017-2024 Albert Astals Cid <aacid@kde.org>
 // Copyright (C) 2005 Jonathan Blandford <jrb@redhat.com>
 // Copyright (C) 2006, 2007 Jeff Muizelaar <jeff@infidigm.net>
 // Copyright (C) 2006 Takashi Iwai <tiwai@suse.de>
@@ -40,14 +40,15 @@
 // Copyright (C) 2018 Klarälvdalens Datakonsult AB, a KDAB Group company, <info@kdab.com>. Work sponsored by the LiMux project of the city of Munich
 // Copyright (C) 2018, 2020 Adam Reichold <adam.reichold@t-online.de>
 // Copyright (C) 2019 Christian Persch <chpe@src.gnome.org>
-// Copyright (C) 2019 Oliver Sander <oliver.sander@tu-dresden.de>
+// Copyright (C) 2019, 2024 Oliver Sander <oliver.sander@tu-dresden.de>
 // Copyright (C) 2020 Kai Pastor <dg0yt@darc.de>
 // Copyright (C) 2021, 2022 Stefan Löffler <st.loeffler@gmail.com>
 // Copyright (C) 2021 sunderme <sunderme@gmx.de>
 // Copyright (C) 2022 Even Rouault <even.rouault@spatialys.com>
 // Copyright (C) 2022 Claes Nästén <pekdon@gmail.com>
-// Copyright (C) 2023 g10 Code GmbH, Author: Sune Stolborg Vuorela <sune@vuorela.dk>
+// Copyright (C) 2023, 2024 g10 Code GmbH, Author: Sune Stolborg Vuorela <sune@vuorela.dk>
 // Copyright (C) 2023 Shivodit Gill <shivodit.gill@gmail.com>
+// Copyright (C) 2024 Keyu Tao <me@taoky.moe>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -434,8 +435,8 @@ GlobalParams::GlobalParams(const char *customPopplerDataDir) : popplerDataDir(cu
     profileCommands = false;
     errQuiet = false;
 
-    cidToUnicodeCache = new CharCodeToUnicodeCache(cidToUnicodeCacheSize);
-    unicodeToUnicodeCache = new CharCodeToUnicodeCache(unicodeToUnicodeCacheSize);
+    cidToUnicodeCache = std::make_unique<CharCodeToUnicodeCache>(cidToUnicodeCacheSize);
+    unicodeToUnicodeCache = std::make_unique<CharCodeToUnicodeCache>(unicodeToUnicodeCacheSize);
     unicodeMapCache = new UnicodeMapCache();
     cMapCache = new CMapCache();
 
@@ -581,8 +582,6 @@ GlobalParams::~GlobalParams()
     delete sysFonts;
     delete textEncoding;
 
-    delete cidToUnicodeCache;
-    delete unicodeToUnicodeCache;
     delete unicodeMapCache;
     delete cMapCache;
 }
@@ -884,25 +883,23 @@ static FcPattern *buildFcPattern(const GfxFont *font, const GooString *base14Nam
 }
 #endif
 
-GooString *GlobalParams::findFontFile(const std::string &fontName)
+std::optional<std::string> GlobalParams::findFontFile(const std::string &fontName)
 {
-    GooString *path = nullptr;
-
     setupBaseFonts(POPPLER_FONTSDIR);
     globalParamsLocker();
     const auto fontFile = fontFiles.find(fontName);
     if (fontFile != fontFiles.end()) {
-        path = new GooString(fontFile->second);
+        return fontFile->second;
     }
 
-    return path;
+    return {};
 }
 
 #if defined(WITH_FONTCONFIGURATION_FONTCONFIG) || defined(WITH_FONTCONFIGURATION_WIN32)
 
 static bool supportedFontForEmbedding(Unicode uChar, const char *filepath, int faceIndex)
 {
-    if (!GooString::endsWith(filepath, ".ttf") && !GooString::endsWith(filepath, ".ttc") && !GooString::endsWith(filepath, ".otf")) {
+    if (!std::string_view(filepath).ends_with(".ttf") && !std::string_view(filepath).ends_with(".ttc") && !std::string_view(filepath).ends_with(".otf")) {
         // for now we only support ttf, ttc, otf fonts
         return false;
     }
@@ -942,7 +939,7 @@ static bool supportedFontForEmbedding(Unicode uChar, const char *filepath, int f
 // not needed for fontconfig
 void GlobalParams::setupBaseFonts(const char *) { }
 
-GooString *GlobalParams::findBase14FontFile(const GooString *base14Name, const GfxFont *font, GooString *substituteFontName)
+std::optional<std::string> GlobalParams::findBase14FontFile(const GooString *base14Name, const GfxFont *font, GooString *substituteFontName)
 {
     SysFontType type;
     int fontNum;
@@ -950,21 +947,21 @@ GooString *GlobalParams::findBase14FontFile(const GooString *base14Name, const G
     return findSystemFontFile(font, &type, &fontNum, substituteFontName, base14Name);
 }
 
-GooString *GlobalParams::findSystemFontFile(const GfxFont *font, SysFontType *type, int *fontNum, GooString *substituteFontName, const GooString *base14Name)
+std::optional<std::string> GlobalParams::findSystemFontFile(const GfxFont *font, SysFontType *type, int *fontNum, GooString *substituteFontName, const GooString *base14Name)
 {
     const SysFontInfo *fi = nullptr;
     FcPattern *p = nullptr;
-    GooString *path = nullptr;
+    std::optional<std::string> path;
     const std::optional<std::string> &fontName = font->getName();
     GooString substituteName;
     if (!fontName) {
-        return nullptr;
+        return {};
     }
 
     globalParamsLocker();
 
     if ((fi = sysFonts->find(*fontName, font->isFixedWidth(), true))) {
-        path = fi->path->copy();
+        path = fi->path->toStr();
         *type = fi->type;
         *fontNum = fi->fontNum;
         substituteName.Set(fi->substituteName->c_str());
@@ -1059,7 +1056,7 @@ GooString *GlobalParams::findSystemFontFile(const GfxFont *font, SysFontType *ty
                     SysFontInfo *sfi = new SysFontInfo(new GooString(*fontName), bold, italic, oblique, font->isFixedWidth(), new GooString((char *)s), *type, *fontNum, substituteName.copy());
                     sysFonts->addFcFont(sfi);
                     fi = sfi;
-                    path = new GooString((char *)s);
+                    path = std::string((char *)s);
                 } else if (!strncasecmp(ext, ".pfa", 4) || !strncasecmp(ext, ".pfb", 4)) {
                     int weight, slant;
                     bool bold = font->isBold();
@@ -1082,7 +1079,7 @@ GooString *GlobalParams::findSystemFontFile(const GfxFont *font, SysFontType *ty
                     SysFontInfo *sfi = new SysFontInfo(new GooString(*fontName), bold, italic, oblique, font->isFixedWidth(), new GooString((char *)s), *type, *fontNum, substituteName.copy());
                     sysFonts->addFcFont(sfi);
                     fi = sfi;
-                    path = new GooString((char *)s);
+                    path = std::string((char *)s);
                 } else {
                     continue;
                 }
@@ -1098,8 +1095,8 @@ GooString *GlobalParams::findSystemFontFile(const GfxFont *font, SysFontType *ty
         }
         FcFontSetDestroy(set);
     }
-    if (path == nullptr && (fi = sysFonts->find(*fontName, font->isFixedWidth(), false))) {
-        path = fi->path->copy();
+    if (!path && (fi = sysFonts->find(*fontName, font->isFixedWidth(), false))) {
+        path = fi->path->toStr();
         *type = fi->type;
         *fontNum = fi->fontNum;
     }
@@ -1163,11 +1160,16 @@ UCharFontSearchResult GlobalParams::findSystemFontFileForUChar(Unicode uChar, co
             int faceIndex = 0;
             FcChar8 *fcFamily = nullptr;
             FcChar8 *fcStyle = nullptr;
+            FcCharSet *fcCharSet = nullptr;
             FcPatternGetString(fontSet->fonts[i], FC_FILE, 0, &fcFilePath);
             FcPatternGetInteger(fontSet->fonts[i], FC_INDEX, 0, &faceIndex);
             FcPatternGetString(fontSet->fonts[i], FC_FAMILY, 0, &fcFamily);
             FcPatternGetString(fontSet->fonts[i], FC_STYLE, 0, &fcStyle);
-            if (!fcFilePath || !fcFamily || !fcStyle) {
+            FcPatternGetCharSet(fontSet->fonts[i], FC_CHARSET, 0, &fcCharSet);
+            if (!fcFilePath || !fcFamily || !fcStyle || !fcCharSet) {
+                continue;
+            }
+            if (!FcCharSetHasChar(fcCharSet, uChar)) {
                 continue;
             }
 
@@ -1184,7 +1186,7 @@ UCharFontSearchResult GlobalParams::findSystemFontFileForUChar(Unicode uChar, co
 #elif defined(WITH_FONTCONFIGURATION_ANDROID)
 // Uses the font file mapping created by GlobalParams::setupBaseFonts
 // to return the path to a base-14 font file
-GooString *GlobalParams::findBase14FontFile(const GooString *base14Name, const GfxFont *font, GooString *substituteFontName)
+std::optional<std::string> GlobalParams::findBase14FontFile(const GooString *base14Name, const GfxFont *font, GooString *substituteFontName)
 {
     return findFontFile(base14Name->toStr());
 }
@@ -1207,13 +1209,12 @@ struct AFontDestroyer
 
 #    endif
 
-GooString *GlobalParams::findSystemFontFile(const GfxFont *font, SysFontType *type, int *fontNum, GooString *substituteFontName, const GooString *base14Name)
+std::optional<std::string> GlobalParams::findSystemFontFile(const GfxFont *font, SysFontType *type, int *fontNum, GooString *substituteFontName, const GooString *base14Name)
 {
-    GooString *path = nullptr;
     const std::optional<std::string> &fontName = font->getName();
 
     if (!fontName) {
-        return nullptr;
+        return {};
     }
 
     globalParamsLocker();
@@ -1238,13 +1239,13 @@ GooString *GlobalParams::findSystemFontFile(const GfxFont *font, SysFontType *ty
 
     // Get font match and the font file's path
     std::unique_ptr<AFont, AFontDestroyer> afont { AFontMatcher_match(fontmatcher.get(), genericFontFamily.c_str(), (uint16_t *)u"A", 1, nullptr) };
-    path = new GooString(AFont_getFontFilePath(afont.get()));
+    std::string path = std::string(AFont_getFontFilePath(afont.get()));
 
     // Set the type of font. Fonts returned by AFontMatcher are of
     // four possible types - ttf, otf, ttc, otc.
-    if (path->endsWith(".ttf") || path->endsWith(".otf")) {
+    if (path.ends_with(".ttf") || path.ends_with(".otf")) {
         *type = sysFontTTF;
-    } else if (path->endsWith(".ttc") || path->endsWith(".otc")) {
+    } else if (path.ends_with(".ttc") || path.ends_with(".otc")) {
         *type = sysFontTTC;
     }
 #    else
@@ -1331,7 +1332,7 @@ UCharFontSearchResult GlobalParams::findSystemFontFileForUChar(Unicode uChar, co
 #elif defined(WITH_FONTCONFIGURATION_WIN32)
 #    include "GlobalParamsWin.cc"
 
-GooString *GlobalParams::findBase14FontFile(const GooString *base14Name, const GfxFont *font, GooString * /*substituteFontName*/)
+std::optional<std::string> GlobalParams::findBase14FontFile(const GooString *base14Name, const GfxFont *font, GooString * /*substituteFontName*/)
 {
     return findFontFile(base14Name->toStr());
 }
@@ -1350,7 +1351,7 @@ UCharFontSearchResult GlobalParams::findSystemFontFileForUChar(Unicode uChar, co
     return {};
 }
 
-GooString *GlobalParams::findBase14FontFile(const GooString *base14Name, const GfxFont *font, GooString * /*substituteFontName*/)
+std::optional<std::string> GlobalParams::findBase14FontFile(const GooString *base14Name, const GfxFont *font, GooString * /*substituteFontName*/)
 {
     return findFontFile(base14Name->toStr());
 }
@@ -1413,20 +1414,19 @@ void GlobalParams::setupBaseFonts(const char *dir)
     }
 }
 
-GooString *GlobalParams::findSystemFontFile(const GfxFont *font, SysFontType *type, int *fontNum, GooString * /*substituteFontName*/, const GooString * /*base14Name*/)
+std::optional<std::string> GlobalParams::findSystemFontFile(const GfxFont *font, SysFontType *type, int *fontNum, GooString * /*substituteFontName*/, const GooString * /*base14Name*/)
 {
     const SysFontInfo *fi;
-    GooString *path;
+    std::optional<std::string> path;
 
     const std::optional<std::string> &fontName = font->getName();
     if (!fontName) {
-        return nullptr;
+        return {};
     }
 
-    path = nullptr;
     globalParamsLocker();
     if ((fi = sysFonts->find(*fontName, font->isFixedWidth(), false))) {
-        path = fi->path->copy();
+        path = fi->path->toStr();
         *type = fi->type;
         *fontNum = fi->fontNum;
     }
@@ -1469,9 +1469,9 @@ bool GlobalParams::getErrQuiet()
     return errQuiet;
 }
 
-CharCodeToUnicode *GlobalParams::getCIDToUnicode(const GooString *collection)
+std::shared_ptr<CharCodeToUnicode> GlobalParams::getCIDToUnicode(const GooString *collection)
 {
-    CharCodeToUnicode *ctu;
+    std::shared_ptr<CharCodeToUnicode> ctu;
 
     globalParamsLocker();
     if (!(ctu = cidToUnicodeCache->getCharCodeToUnicode(collection))) {
